Require Import syntax List String ZArith.
Open Scope Z_scope.

(* Dans le deuxime TD, nous avons dcrit la smantique naturelle du
  petit langage de programmation en utilisant un prdicat inductif.
  Dans le troisime TD, nous avons dcrit la smantique axiomatique
  en utilisant encore une fois un prdicat inductif.
  Dans le quatrime TD, nous avons montr que la smantique axiomatique
  pouvait tre calcule par une fonction rcursive, ce qui permet d'automatiser
  les preuves.
  Dans ce cinquime TD, nous allons dcrire la smantique naturelle avec
  des fonctions rcursives, de faon  avoir  nouveau une faon automatique
  de calculer des preuves d'excution. *)

(* Le problme le plus dur sera de grer les boucles while, qui peuvent
  ne pas terminer.  Or nous avons seulement appris comment programmer des
  fonctions dont le calcul doit toujours terminer.  Pour contourner ce 
  problme, nous nous limiterons aux excutions pour lesquelles le nombre
  d'excution des boucles est limit par un nombre entier fix  l'avance. *)

(* RAPPEL. *)

(* Nous encoderons les environnements par des listes de paires. *)

Definition env := list (string*Z).

(* Voici l'encodage en Coq des rgles d'valuation pour les expressions
  arithmtiques. *)

Inductive aeval : env -> aexpr -> Z -> Prop :=
  ae_int : forall r n, aeval r (anum n) n
| ae_var1 : forall r x v, aeval ((x,v)::r) (avar x) v
| ae_var2 : forall r x y v v' , x <> y -> aeval r (avar x) v' ->
              aeval ((y,v)::r) (avar x) v'
| ae_plus : forall r e1 e2 v1 v2,
              aeval r e1 v1 -> aeval r e2 v2 ->
              aeval r (aplus e1 e2) (v1 + v2).

(* Voici la description du jugement d'valuation des expressions boolennes. *)
Inductive beval : env -> bexpr -> bool -> Prop :=
  blt_true : forall r e1 e2 v1 v2,
    aeval r e1 v1 -> aeval r e2 v2 -> v1 < v2 ->
    beval r (blt e1 e2) true
| blt_false : forall r e1 e2 v1 v2,
    aeval r e1 v1 -> aeval r e2 v2 -> v2 <= v1 ->
    beval r (blt e1 e2) false.

(* Voici la description du jugement de mise  jour. *)

Inductive update : env->string->Z->env->Prop :=
| up1 : forall r x v v', update ((x,v)::r) x v' ((x,v')::r)
| up2 : forall r r' x y v v', update r x v' r' ->
            x <> y -> update ((y,v)::r) x v' ((y,v)::r').

(* Et voici le jugement d'excution des instructions. *)


Inductive exec : env -> instr -> env -> Prop :=
  ex_a : forall r x e v r', aeval r e v -> update r x v r' ->
            exec r (assign x e) r'
| ex_s : forall r1 r2 r3 i1 i2, exec r1 i1 r2 -> exec r2 i2 r3 ->
            exec r1 (sequence i1 i2) r3
| ex_wt : forall r1 r2 r3 b i, beval r1 b true -> exec r1 i r2 ->
     exec r2 (while b i) r3 ->
     exec r1 (while b i) r3
| ex_wf : forall r b i, beval r b false -> exec r (while b i) r.

(* FIN DU RAPPEL. *)

(* La fonction suivante parcourt un environnement pour retourner la
  valeur associe  un nom de variable x, lorsque cette variable est
  dfinie dans l'environnement.  La valeur est optionnelle.

  La fonction string_dec compare deux chaines de caractres. *)

Fixpoint lookup (r : env) (x : string) {struct r} : option Z :=
  match r with
    nil => None
  | (y, v)::tl => 
    if string_dec x y then Some v else lookup tl x
  end.

(* Voici un thorme qui dmontre que cette fonction permet bien
  de construire des preuves de aeval. *)

Lemma lookup_sound :
  forall r x v, lookup r x = Some v -> aeval r (avar x) v.
(* On fait la preuve par rcurrence sur l'environement *)
induction r.
simpl.
intros x v H.
(* l'hypothse est absurde, car elle affirme que deux constructeurs
  diffrents du mme langage sont gaux.  La tactique "discriminate" gre
  ce cas. *)
discriminate.
simpl.
intros x v.
destruct a as [y v'].
(* Pour raisonner sur le comportement de string_dec, il suffit de faire
  une analyse sur son comportement.  Pour cette fonction cela ajoute
  automatiquement une formule logique dans le contexte. *)
case (string_dec x y).
(* Dans le premier cas, on a une hypothse supplmentaire qui dit que
  x est gal  y. *)
intros xeqy sv'eqsv.
(* Il faut rcuprer la valeur de v *)
injection sv'eqsv.
(* Cela donne une deux galit. *)
intros v'eqv.
rewrite xeqy.
rewrite v'eqv.
apply ae_var1.
intros xdify loeqv.
apply ae_var2.
assumption.
apply IHr.
assumption.
Qed.

(* A vous de jouer: crivez une fonction qui calcule la valeur d'une
  expression arithmtique lorsque c'est valeur est dfinie.   La valeur
  est  nouveau optionnelle. *)

Fixpoint af ...


(* Dfinissez maintenant une fonction qui calcule la valeur d'une expression
   boolenne. Le premier argument doit tre l'environnement. *)

Definition bf ...

Lemma bf_sound : forall r b v, bf r b = Some v -> beval r b v.
(* A vous de faire cette preuve. *)
(* Inspirez-vous de la preuve wf_sound qui suit.  En particulier, tudiez
   la faon dont la tactique "case_eq" procde. *)

Admitted.

(* La fonction suivante excute un test exprim par une expression boolenne
  puis excute une fonction de type (env -> option env) et rpte  nouveau
  le procd au plus n fois.  L'excution peut s'arrter plus tt si 
  l'expression boolenne n'a pas de valeur ou si la fonction ne retourne
  pas une bonne valeur. *)

Fixpoint wf (n:nat) (b:bexpr) (f : env -> option env) (r : env) 
  {struct n} : option env :=
  match n with
    0%nat => None
  | S p => 
      match bf r b with
        Some v =>
        if v then match f r with Some r' => wf p b f r' | None => None end
        else Some r
      | None => None
      end
  end.

Lemma wf_sound : forall n r b f i r',
  (forall r r', f r = Some r' -> exec r i r') ->
  wf n b f r = Some r' -> exec r (while b i) r'.
induction n; intros r b f i r' rec.
simpl.
intros ; discriminate.
simpl.
case_eq (bf r b).
intros [ | ].
intros qb.
case_eq (f r).
intros r1 qf q.
apply ex_wt with r1.
apply bf_sound.
assumption.
apply rec.
assumption.
apply IHn with (f:=f).
assumption.
assumption.
intros; discriminate.
intros qb q; injection q; intros; subst r'.
apply ex_wf.
apply bf_sound.
auto.
intros; discriminate.
Qed.

(* C'est votre tour.  Construisez une fonction qui excute une instruction
  dans un environnement donn.  La fonction prend un paramtre numrique
  qui limite le nombre de fois que chaque boucle peut tre excute. *)

Fixpoint run (n:nat) (r : env) (i : instr) {struct i} : option env :=
...

(* Dmontrez que lorsque cette fonction retourne une valeur, celle-ci est
  prvue par la smantique. *)

Lemma run_sound : forall n r i r', run n r i = Some r' -> exec r i r'.

